function output = CalculateClusterHomographies(normalMap,depthZ,K,patchResolution,clusterCount,...
    xyWeight,depthWeight)
	
% Inputs :
% Map with the normal for each pixel
% depthZ - 2D array including the depth value of each pixel
% K - Camera intrinsic parameters matrix
% pathResolution - the size of the patch edge
% clusterCount - The number of clusters to calculate

% Output : 
% Struct contains :
% H : matrix contains homography for each cluster (view point), 
% projectArea : area of of each Cluster
% count : the number of clusters 

% Function description : this code calculate clusterCount centers,
% of 5D vector (x,y,z,nx,ny) that is being weighted, ie : giving more weights
% to specific dimentions, the final calulation is based on kmeans algorithm

visualize = true;
mask = depthZ > 0;

validIndexes = find(mask > 0);
[I J] = find(mask > 0);

I = xyWeight * (2 * (I / size(mask,1)) - 1);
J = xyWeight * (2 * (J / size(mask,2)) - 1);

nx = normalMap(:,:,1);
ny = normalMap(:,:,2);
d = depthZ(mask);
nx = nx(mask);
ny = ny(mask);
d_normalized = depthWeight * (2 * ((d - min(d)) / (max(d) - min(d))) - 1);

a = [d_normalized nx ny I J];
[IDX,C,sumd,D] = kmeans(a,clusterCount);

clusters.d = min(d) + (max(d) - min(d)) * 0.5 * (1 + C(:,1));
clusters.n = [C(:,2) C(:,3) sqrt((1 - C(:,2).^2 - C(:,3).^2))];

if (visualize)
    figure(10)
    subplot(231); imshow(depthZ,[]);
    title('depth')
    subplot(232); imshow(normalMap(:,:,1),[]);
    title('nx')
    subplot(233); imshow(normalMap(:,:,2),[]);
    title('ny')
    subplot(234); hist(d)
    subplot(235); hist(nx)
    subplot(236); hist(ny)
    
    figure(11);
    subplot(121); hist(IDX);
    title('cluster counts')
    subplot(122); plot3(clusters.d,clusters.n(:,1),clusters.n(:,2),'*')
    title('cluster centers')
    grid on
end

%%% visualize clusters
clusterImage = zeros(size(depthZ));
for i = 1:clusterCount
    clusterIndexes = validIndexes(IDX == i);
    clusterImage(clusterIndexes) = i;
end

if (visualize)
    figure(12); imshow(clusterImage,[])
    colormap jet
    colorbar
    title('cluster image');
end

%%% calc cluster homographies
theta = acos(clusters.n(:,3));

wx = -clusters.n(:,2);
wy = +clusters.n(:,1);
wz = zeros(clusterCount,1);

w = zeros(clusterCount,3);
w(:,1) = wx;
w(:,2) = wy;
w(:,3) = wz;
w_norm = sqrt(sum(w.^2,2));

wx = (wx./w_norm).*theta;
wy = (wy./w_norm).*theta;

wx(w_norm==0) = 0;
wy(w_norm==0) = 0;

patchRadius = patchResolution/(2*max(K(1,1),K(2,2)));
patchGrid = [-patchRadius -patchRadius 0 ; patchRadius -patchRadius 0 ; patchRadius patchRadius 0 ;-patchRadius patchRadius 0]';

output.H = zeros(3,3,clusterCount);
output.projectionArea = zeros(1,clusterCount);
output.count = clusterCount;

for i = 1:clusterCount
    %%% calc homography
    R = rodrigues([wx(i) wy(i) 0]');
    t = [0 ; 0; clusters.d(i)];
    patchGridT = R * patchGrid + repmat(t,[1 4]);
    projectedPatch = K * patchGridT;
    projectedPatch = projectedPatch ./ repmat(projectedPatch(3,:),[3 1]);
    output.H(:,:,i) = calcH(patchGrid,projectedPatch);
    
    %%% calc projection area
    %%% projection diagonals
    Ax = projectedPatch(1,1);
    Ay = projectedPatch(2,1);
    Bx = projectedPatch(1,2);
    By = projectedPatch(2,2);
    Cx = projectedPatch(1,3);
    Cy = projectedPatch(2,3);
    Dx = projectedPatch(1,4);
    Dy = projectedPatch(2,4);
    
    %%% 0.5 abs(AC cross BD):
    output.projectionArea(i) = 0.5 * abs((Cx - Ax) * (Dy - By) - (Dx - Bx) * (Cy - Ay));
end

%%%
% if (visualize)
% figure(4)
% for i = 1:clusterCount
%     subplot(3,ceil(clusterCount / 3),i); imshow(output.H(:,:,i) / output.H(3,3,i),[])
% end
% colormap jet
% end

drawnow
%pause
end